from celery import shared_task

from .serializers import Rest_Tracker_Request_Serializer

@shared_task()
def rest_tracker_task(task_data):
    serializer = Rest_Tracker_Request_Serializer(data=task_data)
    if serializer.is_valid():
        serializer.save()
        return serializer.data
    else:
        return serializer.errors




from rest_framework import viewsets, mixins

from .models import Rest_Tracker_Request
from .serializers import Rest_Tracker_Request_Serializer


class Rest_Tracker_View(viewsets.GenericViewSet, mixins.ListModelMixin):
    queryset = Rest_Tracker_Request.objects.all().order_by('created')
    serializer_class = Rest_Tracker_Request_Serializer



from rest_framework import serializers

from .models import Rest_Tracker_Request, Rest_Tracker_Responses, Rest_Tracker_Methods, Rest_Tracker_Schemes, \
    Rest_Tracker_User_Agents, Rest_Tracker_Status_Codes, Rest_Tracker_Urls


class Rest_Tracker_Status_Codes_Serializer(serializers.Serializer):
    id = serializers.IntegerField(required=False)
    created = serializers.DateTimeField(required=False)
    status_code = serializers.IntegerField(null=False)

    class Meta:
        fields = ('id', 'created', 'status_code')
        read_only_fields = ('id', 'created')

    def to_representation(self, instance):
        return instance.status_code

class Rest_Tracker_Methods_Serializer(serializers.Serializer):
    id = serializers.IntegerField(required=False)
    created = serializers.DateTimeField(required=False)
    method = serializers.CharField()

    class Meta:
        fields = ('id', 'created', 'method')
        read_only_fields = ('id', 'created')

    def to_representation(self, instance):
        return instance.method

class Rest_Tracker_Schemes_Serializer(serializers.Serializer):
    id = serializers.IntegerField(required=False)
    created = serializers.DateTimeField(required=False)
    scheme = serializers.CharField()

    class Meta:
        fields = ('id', 'created', 'scheme')
        read_only_fields = ('id', 'created')

    def to_representation(self, instance):
        return instance.scheme

class Rest_Tracker_Urls_Serializer(serializers.Serializer):
    id = serializers.IntegerField(required=False)
    created = serializers.DateTimeField(required=False)
    host = serializers.CharField()
    path = serializers.CharField()
    raw = serializers.CharField()

    class Meta:
        fields = ('id', 'created', 'host', 'path', 'raw')
        read_only_fields = ('id', 'created')

class Rest_Tracker_User_Agents_Serializer(serializers.Serializer):
    id = serializers.IntegerField(required=False)
    created = serializers.DateTimeField(required=False)
    user_agent = serializers.CharField()

    class Meta:
        fields = ('id', 'created', 'user_agent')
        read_only_fields = ('id', 'created')

    def to_representation(self, instance):
        return instance.user_agent

class Rest_Tracker_Responses_Serializer(serializers.Serializer):
    id = serializers.IntegerField(required=False)
    created = serializers.DateTimeField(required=False)
    status_code = Rest_Tracker_Status_Codes_Serializer()
    content_size = serializers.IntegerField()

    class Meta:
        fields = ('id', 'created', 'status_code', 'content_size')
        read_only_fields = ('id', 'created')

class Rest_Tracker_Request_Serializer(serializers.Serializer):
    id = serializers.IntegerField(required=False)
    created = serializers.DateTimeField(required=False)
    method = Rest_Tracker_Methods_Serializer()
    scheme = Rest_Tracker_Schemes_Serializer()
    url = Rest_Tracker_Urls_Serializer()
    user_agent = Rest_Tracker_User_Agents_Serializer()
    responses = Rest_Tracker_Responses_Serializer(many=True, required=False)

    class Meta:
        fields = ('id', 'created', 'method', 'scheme', 'url', 'user_agent', 'responses')
        read_only_fields = ('id', 'created')

    def create(self, validated_data):
        method_obj = Rest_Tracker_Methods.objects.get_or_create(**validated_data.pop('method'))[0]
        scheme_obj = Rest_Tracker_Schemes.objects.get_or_create(**validated_data.pop('scheme'))[0]
        url_obj = Rest_Tracker_Urls.objects.get_or_create(**validated_data.pop('url'))[0]
        user_agent_obj = Rest_Tracker_User_Agents.objects.get_or_create(**validated_data.pop('user_agent'))[0]
        responses = validated_data.pop('responses')
        req_obj = Rest_Tracker_Request.objects.get_or_create(method=method_obj, scheme=scheme_obj, url=url_obj,
                                                             user_agent=user_agent_obj)[0]
        for resp in responses:
            status_code_obj = Rest_Tracker_Status_Codes.objects.get_or_create(**resp.pop('status_code'))
            resp_obj = Rest_Tracker_Responses.objects.create(status_code=status_code_obj, **resp)
            req_obj.responses.add(resp_obj)
        req_obj.save()
        return req_obj



from rest_framework.request import Request
from typing import Callable
from functools import update_wrapper
from threading import Thread

from .tasks import rest_tracker_task

def start_task(request, response):
    response.accepted_renderer, response.accepted_media_type = request.accepted_renderer, request.accepted_media_type
    response.renderer_context = request.parser_context
    response.render()
    task_data = {
        'method': {'method':request.method},
        'scheme': {'scheme':request.scheme},
        'url': {
            'host':request.get_host(),
            'path': request.path,
            'raw': request.get_raw_uri()
        },
        'user_agent': {'user_agent': request.headers.get('User-Agent', 'UNKNOWN')},
        'responses': [{
            'status_code': {'status_code':response.status_code}, 'content_size': len(response.content)
        }]
    }
    rest_tracker_task.delay(task_data)


class tracker(object):
    def __init__(self, function: Callable):
        """initialize aiobulk
        the function being decorated will receive a new attribute for the bulk call
        :param function: function being decorated
        """
        self.__self__ = None
        self.__bound__ = function
        update_wrapper(self, function)

    def __call__(self, *args: tuple, **kwargs: dict):
        """call base function"""
        if self.__self__ is not None:
            args = (self.__self__,) + args
        response = self.__bound__(*args, **kwargs)
        for a in args:
            if isinstance(a, Request):
                Thread(target=start_task, args=(a, response)).start()
                break
        return response

    def __get__(self, instance, _):
        """update self if instance is found
        :param instance: cls instance
        :return: return self with instance
        """
        if instance is None:
            return self
        else:
            self.__self__ = instance
            return self


from django.db import models

class Rest_Tracker_Status_Codes(models.Model):
    created = models.DateTimeField(auto_now_add=True)
    status_code = models.IntegerField(null=False)

    class Meta:
        db_table = 'rest_tracker_status_codes'
        app_label = 'rest_tracker'

class Rest_Tracker_Methods(models.Model):
    created = models.DateTimeField(auto_now_add=True)
    method = models.CharField(null=False, max_length=16)

    class Meta:
        db_table = 'rest_tracker_methods'
        app_label = 'rest_tracker'

class Rest_Tracker_Schemes(models.Model):
    created = models.DateTimeField(auto_now_add=True)
    scheme = models.TextField(null=False)

    class Meta:
        db_table = 'rest_tracker_schemes'
        app_label = 'rest_tracker'

class Rest_Tracker_Urls(models.Model):
    created = models.DateTimeField(auto_now_add=True)
    host = models.TextField(null=False)
    path = models.TextField(null=False)
    raw = models.TextField(null=False)

    class Meta:
        db_table = 'rest_tracker_hosts'
        app_label = 'rest_tracker'

class Rest_Tracker_User_Agents(models.Model):
    created = models.DateTimeField(auto_now_add=True)
    user_agent = models.TextField(null=True)

    class Meta:
        db_table = 'rest_tracker_user_agents'
        app_label = 'rest_tracker'

class Rest_Tracker_Responses(models.Model):
    created = models.DateTimeField(auto_now_add=True)
    status_code = models.ForeignKey(Rest_Tracker_Status_Codes)
    content_size = models.IntegerField(default=0)

    class Meta:
        db_table = 'rest_tracker_response'
        app_label = 'rest_tracker'

class Rest_Tracker_Request(models.Model):
    created = models.DateTimeField(auto_now_add=True)
    method = models.ForeignKey(Rest_Tracker_Methods, on_delete=models.DO_NOTHING)
    scheme = models.ForeignKey(Rest_Tracker_Schemes, on_delete=models.DO_NOTHING)
    url = models.ForeignKey(Rest_Tracker_Urls, on_delete=models.DO_NOTHING)
    user_agent = models.ForeignKey(Rest_Tracker_User_Agents, on_delete=models.DO_NOTHING)
    responses = models.ManyToManyField(Rest_Tracker_Responses)

    class Meta:
        db_table = 'rest_tracker_request'
        app_label = 'rest_tracker'



from rest_framework import routers

from .views import Rest_Tracker_View

router = routers.DefaultRouter(trailing_slash=False)
router.register(r'rest_tracker/', Rest_Tracker_View, basename='tracker')
urlpatterns = router.urls